void FONT_get_size( unsigned char *msg, int *column, int *row, int dual_size );
int FONT_get_bitmap( unsigned char *msg, char *font,
			unsigned char *buf, int buf_width,
			int size_h, int size_v, int add_h, int add_v,
			int dual_size );
void FONT_bit_reorder( unsigned char *dst, unsigned char *src, int len );
int FONT_init( char *vfcap );
int FONT_deinit( void );
